--
-- Mulcher
-- Specialization for vehicles with mulcher functionality
--
-- @author  John Deere 6930
-- @date 25/05/2021
--

Mulcher = {};

Mulcher.modDir = g_currentModDirectory;
Mulcher.MOD_NAME = g_currentModName;

function Mulcher.initSpecialization()
    g_workAreaTypeManager:addWorkAreaType("mulcher", true);
end;

function Mulcher.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(WorkArea, specializations);
end;

function Mulcher.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "processMulcherArea",              Mulcher.processMulcherArea);
    SpecializationUtil.registerFunction(vehicleType, "processDropEffect",               Mulcher.processDropEffect);
    SpecializationUtil.registerFunction(vehicleType, "setDropEffectEnabled",            Mulcher.setDropEffectEnabled);
    SpecializationUtil.registerFunction(vehicleType, "loadFruitTypeConverterFromXML",   Mulcher.loadFruitTypeConverterFromXML);
end;

function Mulcher.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadWorkAreaFromXML", 	Mulcher.loadWorkAreaFromXML);
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "doCheckSpeedLimit",   	Mulcher.doCheckSpeedLimit);
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDirtMultiplier",   	Mulcher.getDirtMultiplier);
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getWearMultiplier",   	Mulcher.getWearMultiplier);
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanToggleTurnedOn",	Mulcher.getCanToggleTurnedOn);
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsWorkAreaActive", 	Mulcher.getIsWorkAreaActive);
end;

function Mulcher.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", 					Mulcher);
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", 					Mulcher);
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", 		Mulcher);
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream",		Mulcher);
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOn", 				Mulcher);
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOff", 				Mulcher);
    SpecializationUtil.registerEventListener(vehicleType, "onStartWorkAreaProcessing", 	Mulcher);
    SpecializationUtil.registerEventListener(vehicleType, "onEndWorkAreaProcessing", 	Mulcher);
end;

function Mulcher:onLoad(savegame)
	if self.spec_mulcher == nil then
		self.spec_mulcher = self[string.format("spec_%s.mulcher", Mulcher.MOD_NAME)]
	end;
    local spec = self.spec_mulcher;
    if self.isClient then
        spec.animationNodes = g_animationManager:loadAnimations(self.xmlFile, "vehicle.mulcher.animationNodes", self.components, self, self.i3dMappings);

        spec.dropEffects = {};
        local i = 0;
        while true do
            local key = string.format("vehicle.mulcher.dropEffects.dropEffect(%d)", i);
            if not hasXMLProperty(self.xmlFile, key) then
                break
            end;
            local effects = g_effectManager:loadEffect(self.xmlFile, key, self.components, self, self.i3dMappings);
            if effects ~= nil then
                local dropEffect = {};
                dropEffect.effects = effects;
                dropEffect.workAreaIndex = getXMLInt(self.xmlFile, key .. "#workAreaIndex");
                dropEffect.activeTime = -1;
                dropEffect.activeTimeDuration = 750;
                dropEffect.isActive = false;
                dropEffect.isActiveSent = false;
                table.insert(spec.dropEffects, dropEffect);
            end
            i = i + 1
        end
    end

	spec.needsActivation = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher.needsActivation#value"), false);
	spec.onlyActiveWhenLowered = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher#onlyActiveWhenLowered"), true);
	spec.prepareFruits = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher#prepareFruits"), true);
	spec.worksInGround = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher#worksInGround"), false);
	spec.convertHeightDensityToSpray = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher#convertHeightDensityToSpray"), true);
	spec.removeWeeds = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher#removeWeeds"), true);
	spec.eraseTireTracks = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher#eraseTireTracks"), false);
	spec.nitrogenAddFactor = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.mulcher#nitrogenAddFactor"), 0.0);
	spec.limitToField = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.mulcher#limitToField"), false);

	spec.fruitTypeConverters = {};

    local useDefaultForFruits = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.mulcher.fruitTypeConverter#useDefaultForFruits"), false);
    if useDefaultForFruits then
        local inputFruitType = getXMLString(self.xmlFile, "vehicle.mulcher.fruitTypeConverter.converter(0)#inputFruitType");
        if inputFruitType == "DEFAULT" then
            local _, data = self:loadFruitTypeConverterFromXML(self.xmlFile, "vehicle.mulcher.fruitTypeConverter.converter(0)", true);
            for fruitType, fruit in pairs(g_currentMission.fruits) do
                if fruitType ~= nil and data ~= nil then
                    if fruitType == FruitType.GRASS then
                        data.isDefaultFruit = false;
                    end;
                    spec.fruitTypeConverters[fruitType] = data;
                end;
            end;
        else
            g_logManager:xmlWarning(self.configFileName, "To use default fruitData 'vehicle.mulcher.fruitTypeConverter.converter(0)#inputFruitType' has to be 'DEFAULT'");
        end;
    end;
    
    local i = 0;
    while true do
        local key = string.format("vehicle.mulcher.fruitTypeConverter.converter(%d)", i)
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end;
        local input, data = self:loadFruitTypeConverterFromXML(self.xmlFile, key, false);
        if input ~= nil and data ~= nil then
            spec.fruitTypeConverters[input] = data;
        end;
        i = i + 1;
    end;
	
    spec.workAreaParameters = {};
    spec.workAreaParameters.lastChangedArea = 0;
    spec.workAreaParameters.lastStatsArea = 0;
    spec.workAreaParameters.lastTotalArea = 0;
    spec.workAreaParameters.lastUsedAreas = 0;
    spec.workAreaParameters.lastUsedAreasSum = 0;
    spec.workAreaParameters.lastUsedAreasPct = 0;
    spec.workAreaParameters.lastUsedAreasTime = 0;

    spec.isWorking = false
    spec.dirtyFlag = self:getNextDirtyFlag()

    if self.addAITerrainDetailRequiredRange ~= nil then
        self:addAITerrainDetailRequiredRange(g_currentMission.cultivatorValue,  g_currentMission.cultivatorValue,   g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels);
        self:addAITerrainDetailRequiredRange(g_currentMission.plowValue,        g_currentMission.plowValue,         g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels);
        self:addAITerrainDetailRequiredRange(g_currentMission.sowingValue,      g_currentMission.sowingValue,       g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels);
        self:addAITerrainDetailRequiredRange(g_currentMission.sowingWidthValue, g_currentMission.sowingWidthValue,  g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels);
        self:addAITerrainDetailRequiredRange(g_currentMission.grassValue,       g_currentMission.grassValue,        g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels);
    end;	
end;

function Mulcher:onDelete()
    if self.isClient then
        local spec = self.spec_mulcher
        for _, dropEffect in pairs(spec.dropEffects) do
            g_effectManager:deleteEffects(dropEffect.effects);
        end;
        g_animationManager:deleteAnimations(spec.animationNodes);
    end;
end;

function Mulcher:onReadUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        local spec = self.spec_mulcher;
        if #spec.dropEffects > 0 then
            if streamReadBool(streamId) then
                for _, dropEffect in ipairs(spec.dropEffects) do
                    dropEffect.fillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS);
                    self:setDropEffectEnabled(dropEffect, streamReadBool(streamId));
                end;
            end;
        end;
    end;
end;

function Mulcher:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        local spec = self.spec_mulcher;
        if #spec.dropEffects > 0 then
            if streamWriteBool(streamId, bitAND(dirtyMask, spec.dirtyFlag) ~= 0) then
                for _, dropEffect in ipairs(spec.dropEffects) do
                    streamWriteUIntN(streamId, dropEffect.fillType or FillType.UNKNOWN, FillTypeManager.SEND_NUM_BITS);
                    streamWriteBool(streamId, dropEffect.isActiveSent);
                end;
            end;
        end;
    end;
end;

function Mulcher:getCanToggleTurnedOn(superFunc)
    local spec = self.spec_mulcher;
    if not spec.needsActivation then
        return false;
    end;
    return superFunc(self);
end;

function Mulcher:getIsWorkAreaActive(superFunc, workArea)
    if workArea.type == WorkAreaType.MULCHER then
        local spec = self.spec_mulcher;
        if spec.needsActivation then
			if self.getIsTurnedOn ~= nil then
				if not self:getIsTurnedOn() then
					return false;
				end;
			end;
        end;
		if spec.onlyActiveWhenLowered and self.getIsLowered ~= nil then
            if not self:getIsLowered(false) then
                return false;
            end;
        end;
    end;
    return superFunc(self, workArea)
end;

function Mulcher:processMulcherArea(workArea, dt)
    local spec = self.spec_mulcher;
	
    if self:getLastSpeed() > 1 and (not spec.needsActivation or (self.getIsTurnedOn ~= nil and self:getIsTurnedOn())) then
        spec.isWorking = true;
    end;

    local xs,_,zs = getWorldTranslation(workArea.start);
    local xw,_,zw = getWorldTranslation(workArea.width);
    local xh,_,zh = getWorldTranslation(workArea.height);
    local performedDefaultFruit = false;
	for inputFruitType, converterData in pairs(spec.fruitTypeConverters) do
        local doMulch = true;
        
        --performance optimization
        if converterData.isDefaultFruit then
            doMulch = not performedDefaultFruit;
            if not performedDefaultFruit then
                performedDefaultFruit = true;
            end;
        end;
        
        -- do minimalized function
        if doMulch then
            local aiInUse = self:getIsAIActive();
            local changedArea = MulcherDensityUtil.updateMulcherArea(inputFruitType, xs,zs, xw,zw, xh,zh, converterData.setSprayLevel, converterData.setChoppedStraw, spec.convertHeightDensityToSpray, spec.worksInGround, spec.prepareFruits, spec.removeWeeds, spec.eraseTireTracks, spec.nitrogenAddFactor, aiInUse or spec.limitToField);
            if changedArea > 0 then
                local pixelToSqm = g_currentMission:getFruitPixelsToSqm();
                local sqm = changedArea * pixelToSqm;
                local litersToDrop = sqm * g_fruitTypeManager:getFillTypeLiterPerSqm(inputFruitType, 1);
                workArea.lastPickupLiters = litersToDrop;
                workArea.litersToDrop = workArea.litersToDrop + litersToDrop;

                spec.workAreaParameters.lastChangedArea = spec.workAreaParameters.lastChangedArea + changedArea;
                spec.workAreaParameters.lastStatsArea = spec.workAreaParameters.lastStatsArea + changedArea;
                spec.workAreaParameters.lastTotalArea   = spec.workAreaParameters.lastTotalArea + changedArea;
                spec.workAreaParameters.lastUsedAreas = spec.workAreaParameters.lastUsedAreas + 1;
            end;
        end;
    end;
    spec.workAreaParameters.lastUsedAreasSum = spec.workAreaParameters.lastUsedAreasSum + 1;
    return spec.workAreaParameters.lastChangedArea, spec.workAreaParameters.lastTotalArea;
end;

function Mulcher:processDropEffect(workArea, dt)
    local spec = self.spec_mulcher;
	if workArea.litersToDrop > 0 then
		self:raiseActive();
	end;
    for _, dropEffect in pairs(spec.dropEffects) do
        if workArea.litersToDrop > 0 and (dropEffect.workAreaIndex == nil or dropEffect.workAreaIndex == workArea.workAreaIndex) then
            if dropEffect.fillType ~= workArea.fillType then
                dropEffect.fillType = workArea.fillType
                g_effectManager:setFillType(dropEffect.effects, dropEffect.fillType)
            end
            dropEffect.activeTime = dropEffect.activeTimeDuration
            dropEffect.isActive = true
        else
            dropEffect.activeTime = math.max(dropEffect.activeTime - dt, 0)
            if dropEffect.activeTime > 0 then
                dropEffect.isActive = true
            end
        end
    end
end;

function Mulcher:setDropEffectEnabled(dropEffect, isActive)
    dropEffect.isActive = isActive;
    if self.isClient then
        if isActive then
            g_effectManager:setFillType(dropEffect.effects, dropEffect.fillType);
            g_effectManager:startEffects(dropEffect.effects);
        else
            g_effectManager:stopEffects(dropEffect.effects);
        end;
    end;
end;

function Mulcher:loadWorkAreaFromXML(superFunc, workArea, xmlFile, key)
    local retValue = superFunc(self, workArea, xmlFile, key);
    if workArea.type == WorkAreaType.DEFAULT then
        workArea.type = WorkAreaType.MULCHER;
    end;	
    if workArea.type == WorkAreaType.MULCHER then
        workArea.dropWindrow = Utils.getNoNil(getXMLBool(xmlFile, key .. ".mulcher#dropWindrow"), true);
        workArea.lastPickupLiters = 0;
        workArea.litersToDrop = 0;
    end;
    return retValue;
end;

function Mulcher:doCheckSpeedLimit(superFunc)
    local spec = self.spec_mulcher;
    return superFunc(self) or ((not spec.needsActivation or self:getIsTurnedOn()) and (self.getIsLowered == nil or self:getIsLowered()));
end;

function Mulcher:getDirtMultiplier(superFunc)
    local spec = self.spec_mulcher;
    local multiplier = superFunc(self);
    if spec.isWorking then
        multiplier = multiplier + self:getWorkDirtMultiplier() * self:getLastSpeed() / self.speedLimit;
    end;
    return multiplier;
end;

function Mulcher:getWearMultiplier(superFunc)
    local spec = self.spec_mulcher;
    local multiplier = superFunc(self);
    if spec.isWorking then
        multiplier = multiplier + self:getWorkWearMultiplier() * self:getLastSpeed() / self.speedLimit;
    end;
    return multiplier;
end;

function Mulcher:onTurnedOn()
    if self.isClient then
        local spec = self.spec_mulcher;
        g_animationManager:startAnimations(spec.animationNodes);
    end;
end;

function Mulcher:onTurnedOff()
    if self.isClient then
        local spec = self.spec_mulcher;
        for _, dropEffect in pairs(spec.dropEffects) do
            self:setDropEffectEnabled(dropEffect, false);
        end;
        g_animationManager:stopAnimations(spec.animationNodes);
    end;
end;

function Mulcher:onStartWorkAreaProcessing(dt)
    local spec = self.spec_mulcher;
    if self.isServer then
        for _, dropEffect in pairs(spec.dropEffects) do
            if dropEffect.isActive ~= dropEffect.isActiveSent then
                dropEffect.isActiveSent = dropEffect.isActive;
                self:setDropEffectEnabled(dropEffect, dropEffect.isActiveSent);
                self:raiseDirtyFlags(spec.dirtyFlag);
            end;

            dropEffect.isActive = false;
        end;
    end;
    spec.workAreaParameters.lastChangedArea = 0;
    spec.workAreaParameters.lastStatsArea = 0;
    spec.workAreaParameters.lastTotalArea = 0;
    spec.isWorking = false;
end;

function Mulcher:onEndWorkAreaProcessing(dt, hasProcessed)
    local spec = self.spec_mulcher
    for _, workArea in ipairs(self.spec_workArea.workAreas) do
        if workArea.litersToDrop > 0 then
            self:processDropEffect(workArea, dt);
        end;
		workArea.litersToDrop = math.max(0, workArea.litersToDrop - workArea.lastPickupLiters);
    end;
	
    if self.isServer then
        local lastStatsArea = spec.workAreaParameters.lastStatsArea;
        local lastTotalArea = spec.workAreaParameters.lastTotalArea;
        if lastStatsArea > 0 then
            local ha = MathUtil.areaToHa(lastStatsArea, g_currentMission:getFruitPixelsToSqm()); -- 4096px are mapped to 2048m
            
            local lastTouchedFarmlandFarmId = self:getLastTouchedFarmlandFarmId();
            if lastTouchedFarmlandFarmId == 0 then
                lastTouchedFarmlandFarmId = self:getActiveFarm()
            end;
            local stats = g_currentMission:farmStats(lastTouchedFarmlandFarmId);
            stats:updateStats("workedHectares", ha);
        end;
    end;
    spec.workAreaParameters.lastUsedAreasTime = spec.workAreaParameters.lastUsedAreasTime + dt;
    if spec.workAreaParameters.lastUsedAreasTime > 500 then
        spec.workAreaParameters.lastUsedAreasPct = spec.workAreaParameters.lastUsedAreas / math.max(spec.workAreaParameters.lastUsedAreasSum, 0.01);
        spec.workAreaParameters.lastUsedAreas = 0;
        spec.workAreaParameters.lastUsedAreasSum = 0;
        spec.workAreaParameters.lastUsedAreasTime = 0;
    end;
end;

-- load converter data from xml
function Mulcher:loadFruitTypeConverterFromXML(xmlFile, key, isDefaultFruit)
    local inputIndex = nil;
    local data = nil;
    local inputFruitType = getXMLString(xmlFile, key .. "#inputFruitType");
    if inputFruitType ~= nil and inputFruitType ~= "" then
        inputIndex = g_fruitTypeManager.nameToIndex[inputFruitType];
		if inputIndex ~= nil or isDefaultFruit then
			data = {};
            data.isDefaultFruit = Utils.getNoNil(isDefaultFruit, false);
			data.setSprayLevel = Utils.getNoNil(getXMLBool(xmlFile, key .. "#setSprayLevel"), false);
			data.setChoppedStraw = Utils.getNoNil(getXMLBool(xmlFile, key .. "#setChoppedStraw"), false);
		end;
    end;
    return inputIndex, data;
end;

